/*
 * Last Modified: 2009-08-11
 * Last Modified By: Alex Nisenboim
 * Original Author: Alex Nisenboim
 *
 */

/*
 * The contents of this file are subject to the AOLserver Public License
 * Version 1.1 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License at
 * http://aolserver.com/.
 *
 * Software distributed under the License is distributed on an "AS IS"
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See
 * the License for the specific language governing rights and limitations
 * under the License.
 *
 * The Original Code is AOLserver Code and related documentation
 * distributed by AOL.
 * 
 * The Initial Developer of the Original Code is America Online,
 * Inc. Portions created by AOL are Copyright (C) 1999 America Online,
 * Inc. All Rights Reserved.
 *
 * Alternatively, the contents of this file may be used under the terms
 * of the GNU General Public License (the "GPL"), in which case the
 * provisions of GPL are applicable instead of those above.  If you wish
 * to allow use of your version of this file only under the terms of the
 * GPL and not to allow others to use your version of this file under the
 * License, indicate your decision by deleting the provisions above and
 * replace them with the notice and other provisions required by the GPL.
 * If you do not delete the provisions above, a recipient may use your
 * version of this file under either the License or the GPL.
 */

#include "ns.h"

#define ENABLE_COMPRESSION  1
#define DISABLE_COMPRESSION 0

typedef struct _GzFileFilterConfig {
	char        *extension;
	char        *mode;
} GzFileFilterConfig;

typedef GzFileFilterConfig* ptrGzFileFilterConfig;


static int returnGzipFile (Ns_Conn *conn, char *extension, char *mode);
static int GzFile (char *ifile, char *extension);
static int gzFileProcFilter(void *arg, Ns_Conn *conn, int why);
static int RegisterGzFilter(Tcl_Interp* interp, int when, int objc, Tcl_Obj *CONST objv[]);
static int returnFile (Ns_Conn *conn, char *filename, char* gzfilename);
static int fileExists ( char *filename );
static void setFlag (Ns_Conn *conn, char *mode );
static int AmPoolsCmds(Tcl_Interp *interp, void *ignored);
static int returnGzipFileCmd(ClientData ignored, Tcl_Interp *interp, int objc, Tcl_Obj * CONST objv[]);
static int registerFilterErrorObjCmd(ClientData ignored, Tcl_Interp *interp, int objc, Tcl_Obj *CONST objv[]);
static int am_conn_compressCmd(ClientData ignored, Tcl_Interp *interp, int objc, Tcl_Obj * CONST objv[]);
static int ProcFilterError(void *arg, Ns_Conn *conn, int why);
static int registerErrorFilter(Tcl_Interp* interp, int when, int objc, Tcl_Obj *CONST objv[]);

int Ns_ModuleVersion = 1;
/*
 *----------------------------------------------------------------------
 *
 * AmPools_ModInit--
 *
 *      Tcl load-command entry point.
 *
 * Results:
 *      TCL_OK.
 *
 * Side effects:
 *	
 *
 *----------------------------------------------------------------------
 */
int
AmPools_ModInit(char *server, char *hModule) 
{
    return(Ns_TclInitInterps(server, AmPoolsCmds, NULL));
}

/*
 *----------------------------------------------------------------------
 *
 * AmPoolsCmds--
 *
 *      Create Tcl commands.
 *
 * Results:
 *      TCL_OK.
 *
 * Side effects:
 *	Adds the am_register_gzFileFilter command.
 *
 *----------------------------------------------------------------------
 */
int
AmPoolsCmds(Tcl_Interp *interp, void *ignored)
{
	Tcl_CreateObjCommand(interp, "am_registerGzFileFilter", returnGzipFileCmd, NULL, NULL);
	Tcl_CreateObjCommand(interp, "am_conn_compress", am_conn_compressCmd, NULL, NULL);
	Tcl_CreateObjCommand(interp, "ns_register_filter_error", registerFilterErrorObjCmd, NULL, NULL);
	
	return TCL_OK;
}

/*
 *----------------------------------------------------------------------
 *
 * returnGzipFileCmd --
 *
 *	Implements am_register_gzFileFilter. 
 *
 * Results:
 *	Tcl result. 
 *
 * Side effects:
 *	See docs. 
 *
 *----------------------------------------------------------------------
 */

int
returnGzipFileCmd(ClientData ignored, Tcl_Interp *interp, int objc, Tcl_Obj *CONST objv[])
{
    int       lobjc;
    Tcl_Obj **lobjv;
    int       when, i;
    static CONST char *wopt[] = {
        "prequeue", "preauth", "postauth", "trace", NULL
    };
    enum {
        PQIdx, PRIdx, POIdx, TRIdx,
    } widx;

    if (objc < 2) {
        Tcl_WrongNumArgs(interp, 1, objv, "option ?args?");
        return TCL_ERROR;
    }

	if (objc < 4 || objc > 6) {
        Tcl_WrongNumArgs(interp, 1, objv, "when method url ?extension? ?mode?");
        return TCL_ERROR;
    }
    if (Tcl_ListObjGetElements(interp, objv[1], &lobjc, &lobjv) != TCL_OK) {
		return TCL_ERROR;
    }

    when = 0;
    for (i = 0; i < lobjc; ++i) {
    	if (Tcl_GetIndexFromObj(interp, lobjv[i], wopt, "when", 0, (int *) &widx) != TCL_OK) {
            return TCL_ERROR;
    	}
		switch (widx) {
			case PQIdx:
					when |= NS_FILTER_PRE_QUEUE;
					break;
			case PRIdx:
					when |= NS_FILTER_PRE_AUTH;
					break;
			case POIdx:
					when |= NS_FILTER_POST_AUTH;
					break;
			case TRIdx:
					when |= NS_FILTER_TRACE;
					break;
		}
    }
    if (when == 0) {
		Tcl_SetResult(interp, "blank filter when specification", TCL_STATIC);
		return TCL_ERROR;
    }

    return RegisterGzFilter(interp, when, objc - 2, objv + 2);
}

/*
 *----------------------------------------------------------------------
 *
 * RegisterGzFilter --
 *
 *	Register Gz filter. 
 *
 * Results:
 *	TCL_ERROR if no server, TCL_OK otherwise. 
 *
 * Side effects:
 *	Will register a filter to run later.
 *
 *----------------------------------------------------------------------
 */

static int
RegisterGzFilter(Tcl_Interp* interp, int when, int objc, Tcl_Obj *CONST objv[])
{
    char *server, *method, *pattern, *extension, *mode;
	ptrGzFileFilterConfig pConfig = NULL;

    server = Ns_TclInterpServer(interp);
	if (!server) {
		return TCL_ERROR;
	}
    method = Tcl_GetString(objv[0]);
	pattern = Tcl_GetString(objv[1]);
    
	pConfig = ns_malloc(sizeof(GzFileFilterConfig));
    extension = (objc > 2 ? Tcl_GetString(objv[2]) : NULL);
	pConfig->extension = (extension ? ns_strdup(extension) : NULL);

    mode = (objc > 3 ? Tcl_GetString(objv[3]) : NULL);
	pConfig->mode = (mode ? ns_strdup(mode) : NULL);

	if (pConfig->extension || pConfig->mode) {	
		Ns_RegisterFilter(server, method, pattern, gzFileProcFilter, when|NS_FILTER_FIRST, pConfig);
	}

    return TCL_OK;
}
/*
 *----------------------------------------------------------------------
 *
 * gzFileProcFilter --
 *
 *	Callback for C-based returnGzipFile filters.
 *
 * Results:
 *	Standard filter result.
 *
 * Side effects:
 *	None. 
 *
 *----------------------------------------------------------------------
 */

static int
gzFileProcFilter(void *arg, Ns_Conn *conn, int why)
{
    ptrGzFileFilterConfig pConfig = arg;

	return returnGzipFile(conn, pConfig->extension, pConfig->mode);
}

/*
 *----------------------------------------------------------------------
 *
 * returnGzipFile --
 *
 *	Callback for C-based returnGzipFile filters.
 *
 * Results:
 *	Standard filter result.
 *
 * Side effects:
 *	None. 
 *
 *----------------------------------------------------------------------
 */

static
int returnGzipFile (Ns_Conn *conn, char *extension, char *mode) {

	char *acceptEncoding = NULL, *url = NULL;
	Ns_DString file, gzfile;
    int status = NS_OK;
    char *server = Ns_ConnServer(conn);
    Ns_Set *headers = Ns_ConnHeaders(conn);

	if(!headers){
	    return status;
	}
	
	if (!(acceptEncoding = Ns_SetIGet(headers, "Accept-Encoding"))){
		return status;
	}
	
	if (!strstr(acceptEncoding, "gzip")) {
		return status;
	}
	if(extension && strcmp("", extension)) { 
	    // if compressed file extension is given
	    // build the expected name of compressed file 
		
		url = Ns_ConnUrl(conn);               
	    Ns_DStringInit(&file);
		Ns_UrlToFile(&file, server, url);
		Ns_DStringInit(&gzfile);
		Ns_DStringAppend(&gzfile, file.string);
		Ns_DStringAppend(&gzfile, "."); 
		Ns_DStringAppend(&gzfile, extension); 
		if ( fileExists(gzfile.string) ){
		     // if the file exists, return it
			returnFile(conn, file.string, gzfile.string);
			status = NS_FILTER_RETURN;
	     } else {
			 // if the file does not exist, check the mode and 
			 // disable/enable compression if the mode is specified
			 setFlag(conn, mode); 
		 }
		 Ns_DStringFree(&gzfile);
         Ns_DStringFree(&file);
	 } else {
		 // if the extension is not specified we check
		 // the mode and disable/enable compression if
	     // the mode is specified
		 setFlag(conn, mode);
	 }

	return status;
}

/*
 *----------------------------------------------------------------------
 *
 * returnFile --
 *
 *      .
 *
 * Results:
 *      0/1.
 *
 * Side effects:
 *      None.
 *
 *----------------------------------------------------------------------
 */
static
int returnFile (Ns_Conn *conn, char *filename, char* gzfilename) {

	Ns_Set *outputheaders = Ns_ConnOutputHeaders(conn);
	Ns_SetUpdate(outputheaders, "Content-Encoding", "gzip");
	//prevent double gzipping
	Ns_ConnSetGzipFlag(conn, DISABLE_COMPRESSION);
	return(Ns_ConnReturnFile(conn, 200, Ns_GetMimeType(filename), gzfilename));
}

/*
 *----------------------------------------------------------------------
 *
 * fileExists --
 *
 *      simply checks whether the input file exists or not.
 *
 * Results:
 *      0/1.
 *
 * Side effects:
 *      None.
 *
 *----------------------------------------------------------------------
 */

static
int fileExists ( char *filename ) {

	return(_access(filename, R_OK) == 0 ? 1 : 0);
}


/*
 *----------------------------------------------------------------------
 *
 * setFlag --
 *
 *      enables/disables compression 
 *
 * Results:
 *
 * Side effects:
 *      None.
 *
 *----------------------------------------------------------------------
 */

static
void setFlag (Ns_Conn *conn, char *mode ) {

	if(mode) {
		if(0 == strcmp("none", mode)){
			Ns_ConnSetGzipFlag(conn, DISABLE_COMPRESSION);
		} else {
			Ns_ConnSetGzipFlag(conn, ENABLE_COMPRESSION);
		}
	}
}

/*
 *----------------------------------------------------------------------
 *
 * am_conn_compressCmd --
 *
 *	Implements am_conn_compress. 
 *
 * Results:
 *	Tcl result. 
 *
 * Side effects:
 *	Sets compression enable/disable parameter. 
 *
 *----------------------------------------------------------------------
 */
static 
int am_conn_compressCmd(ClientData ignored, Tcl_Interp *interp, int objc, Tcl_Obj * CONST objv[]){

	int compressFlag;
	Ns_Conn *conn;

	if (objc != 2) {
        Tcl_WrongNumArgs(interp, 1, objv, "compressFlag");
        return TCL_ERROR;
    }

	if (Tcl_GetIntFromObj(interp, objv[1], &compressFlag) != TCL_OK) {    
		return TCL_ERROR;
	}

	if (NULL == (conn = Ns_GetConn())){
		return TCL_ERROR;
	}

	Ns_Log(Debug, "am_conn_compress: setting gzFlag to %d", compressFlag);
	Ns_ConnSetGzipFlag(conn, compressFlag);

    return TCL_OK;
}

/*
 *----------------------------------------------------------------------
 *
 * registerFilterErrorObjCmd --
 *
 *	Implements ns_register_filter_error. 
 *
 * Results:
 *	Tcl result. 
 *
 * Side effects:
 *	Filter that match when|method|urlPattern. 
 *
 *----------------------------------------------------------------------
 */
int
registerFilterErrorObjCmd(ClientData ignored, Tcl_Interp *interp, int objc, Tcl_Obj *CONST objv[])
{
    int       lobjc;
    Tcl_Obj **lobjv;
    int       when, i, len, objIdx = 1;
	char     *option;
    static CONST char *wopt[] = {
        "prequeue", "preauth", "postauth", "trace", NULL
    };
    enum {
        PQIdx, PRIdx, POIdx, TRIdx,
    } widx;

	if (objc < 4 || objc > 5) {
        Tcl_WrongNumArgs(interp, 1, objv, "?-first? when method url");
        return TCL_ERROR;
    }
    
	if (objc == 5){
		option = Tcl_GetStringFromObj(objv[1], &len);
		if (strcmp("-first", option)){
			Ns_Log(Error,"the option should be -first");
			return TCL_ERROR;
		}
        objIdx = 2;
	}


    if (Tcl_ListObjGetElements(interp, objv[objIdx], &lobjc, &lobjv) != TCL_OK) {
		return TCL_ERROR;
    }

    when = 0;
    for (i = 0; i < lobjc; ++i) {
    	if (Tcl_GetIndexFromObj(interp, lobjv[i], wopt, "when", 0, (int *) &widx) != TCL_OK) {
            return TCL_ERROR;
    	}
		switch (widx) {
			case PQIdx:
					when |= NS_FILTER_PRE_QUEUE;
					break;
			case PRIdx:
					when |= NS_FILTER_PRE_AUTH;
					break;
			case POIdx:
					when |= NS_FILTER_POST_AUTH;
					break;
			case TRIdx:
					when |= NS_FILTER_TRACE;
					break;
		}
    }
    if (when == 0) {
		Tcl_SetResult(interp, "blank filter when specification", TCL_STATIC);
		return TCL_ERROR;
    }

	if(objIdx == 2){
		when = when|NS_FILTER_FIRST;
	}

	return registerErrorFilter(interp, when, objc - 2 - (objIdx - 1) , objv + 2 + (objIdx - 1));
}

/*
 *----------------------------------------------------------------------
 *
 * registerErrorFilter --
 *
 *	Register error filter. 
 *
 * Results:
 *	TCL_ERROR if no server, TCL_OK otherwise. 
 *
 * Side effects:
 *	Will register a filter to run later.
 *
 *----------------------------------------------------------------------
 */
static int
registerErrorFilter(Tcl_Interp* interp, int when, int objc, Tcl_Obj *CONST objv[])
{
    char *server, *method, *pattern;

    server = Ns_TclInterpServer(interp);
	if (!server) {
		return TCL_ERROR;
	}
    method = Tcl_GetString(objv[0]);
	pattern = Tcl_GetString(objv[1]);
	Ns_Log(Debug,"when is %d, method is %s, pattern is %s", when, method, pattern);
	Ns_RegisterFilter(server, method, pattern, ProcFilterError, when, "procFilterError");

    return TCL_OK;
}

/*
 *----------------------------------------------------------------------
 *
 * ProcFilterError --
 *
 *	The callback for Tcl error filters. 
 *
 * Results:
 *	NS_FILTER_BREAK.
 *
 * Side effects:
 *	No other filters will run.
 *
 *----------------------------------------------------------------------
 */
static int
ProcFilterError(void *arg, Ns_Conn *conn, int why)
{
    return NS_ERROR;
}
