###############  ampools for AOLserver 4.5.2+  ##################
# Setup optional conn pools and configure their limits.
# Also configure gzip compression options for them
# plus configure skipping of specified filters
# presumably to eliminate the need to allocate
# a Tcl interpreter where not needed or
# to prevent access logging.
#
# Portions based on ideas from Tom Jackson's version of the nszlib module
#
# See http://aolserver.am.net/code/modules/ampools.adpx
#
 
proc ampools_init { } {
    
    set server [ns_info server]
    set cfgsection "ns/server/$server"
    set poolSection $cfgsection/pools
    set poolSet [ns_configsection $poolSection]
    set gzipenabled [ns_config -bool $cfgsection gzip 1]
    set gzipall [ns_config -bool $cfgsection gzipall 0]
    
    if {"$poolSet" ne ""} {
    
        set poolSize [ns_set size $poolSet]
        for {set i 0} {$i < $poolSize} {incr i} {
            set poolName [ns_set key $poolSet $i]
            set poolDescription [ns_set value $poolSet $i]
            set poolConfigSection "ns/server/$server/pool/$poolName"
            set poolConfigSet [ns_configsection $poolConfigSection]
            if {"$poolConfigSet" eq ""} {
                continue
            }
            set poolMinthreads [ns_config $poolConfigSection minthreads 0]
            set poolMaxthreads [ns_config $poolConfigSection maxthreads 10]
            set poolMaxconns   [ns_config $poolConfigSection maxconns [ns_config $poolConfigSection ConnsPerThread 0]] 
            set poolTimeout    [ns_config $poolConfigSection threadtimeout 120]
            
            ns_pools set $poolName -minthreads $poolMinthreads -maxthreads $poolMaxthreads -maxconns $poolMaxconns -timeout $poolTimeout
            ns_log Notice  "$poolName Pool ($poolDescription): [ns_pools get $poolName]"
            
            if {[set maxupload [ns_config $poolConfigSection maxupload]] ne ""} {
                ns_limits set $poolName -maxupload $maxupload
            }
            set maxwait         [ns_config $poolConfigSection MaxWait 100]
            set maxwaittime     [ns_config $poolConfigSection MaxWaitTime 60]
            ns_limits set $poolName -maxrun $poolMaxthreads -maxwait $maxwait -timeout $maxwaittime
            ns_log notice "$poolName limits: [ns_limits get $poolName]"
            
            set poolConfigSize [ns_set size $poolConfigSet]
            set whens [ns_config $poolConfigSection skipfilters]
            for {set j 0} {$j < $poolConfigSize} {incr j} {
                if {[string tolower [ns_set key $poolConfigSet $j]] eq "map"} {
                    set mapList   [ns_set value $poolConfigSet $j]
                    set method          [lindex $mapList 0]
                    set url             [lindex $mapList 1]
                    set extension       [lindex $mapList 2]
                    set compressionMode [lindex $mapList 3]
                    
                    # Register the method and url to this pool and limit-set:
                    ns_pools register $poolName $server $method $url
                    ns_limits register $poolName $server $method $url
                    ns_log Notice "registering url map for pool $poolName in server $server: $method $url"
                    
                    if {[string match *\\** $url]} {
                        set pattern $url
                    } else {
                        set pattern "[string trimright $url /]/*"
                    }
                    
                    set skipTrace 0
                    foreach when $whens {
                        if {$when == "nslog" || $when == "void"} {
                            # register a filter that retuns an error as a trace filter
                            # to prevent firing of the void_trace filters and nslog
                            if {[lsearch $whens "trace"] == -1} {
                                ns_register_filter_error trace $method $pattern
                                ns_log Notice "disabling void filters and nslog for: $method $pattern"
                            } else {
                                # make sure we get placed ahead of the trace shortcut filter
                                ns_register_filter_error -first trace $method $pattern
                                set skipTrace 1
                                ns_log Notice "disabling trace filters for: $method $pattern"
                                ns_log Notice "disabling void filters and nslog for: $method $pattern"
                            }
                            continue
                        } elseif {$when == "trace" && $skipTrace} {
                            continue
                        }
                        ns_register_filter_shortcut $when $method $pattern
                        ns_log Notice "disabling $when filters for: $method $pattern"
                    }
                    
                    # this MUST occur after the shortcut registration to get fired before the shortcut:
                    if {$gzipenabled} {
                        am_registerGzFileFilter preauth $method $pattern $extension $compressionMode
                        ns_log Notice "Compression settings set for $method $pattern: pre-compressed extension = $extension; on-the-fly compression = $compressionMode"
                    }
                }
            }
        }
    }
        
    if {$gzipall} {
        am_registerGzFileFilter preauth GET /* "" gzip
        am_registerGzFileFilter preauth POST /* "" gzip
        ns_log Notice "Enabled gzip compression by default on all GET and POST requests"
    }
}

if [catch {
    ns_log Notice "ampools init..."
    ampools_init
    rename ampools_init {}
} errMsg] {
    ns_log Error "Cannot init ampools on [ns_info server]. Error messages: $errMsg"
}


